/* $Id: csg.cc,v 1.6 1997/04/12 22:32:46 sean Exp sean $ */
/*
$Log: csg.cc,v $
Revision 1.6  1997/04/12 22:32:46  sean
*** empty log message ***

Revision 1.5  1995/08/03 18:45:43  sean
used purify to fix memory leak, need to add bd. spheres to csg nodes.

*/
#include "globals.h"
#include "csg.h"

CSGnode::CSGnode(csg_op_tag otag, csg_node_tag sltag, csg_node_tag srtag,
		 void *lbranch, void *rbranch)
{
  op_tag = otag;
  ltag = sltag;
  rtag = srtag;
  if (ltag == csg_prim)
    surf_left = (Surface *) lbranch;
  else
    csg_left = (CSGnode *) lbranch;
  if (op_tag != csg_op_not)
    if (rtag == csg_prim)
      surf_right = (Surface *) rbranch;
    else
      csg_right = (CSGnode *) rbranch;
}

void CSGnode::print(int level)
{
  int i;

  for (i = 0; i < level; i++)
    printf(" ");
  switch (op_tag)
    {
    case csg_op_union: printf("+\n");
      break;
    case csg_op_intersection: printf("*\n");
      break;
    case csg_op_not: printf("!\n");
      break;
    case csg_op_is: printf("=\n");
      break;
    default:
      break;
    } //switch
  if (ltag == csg_prim)
    surf_left->Print();
  else
    csg_left->print(level + 1);
  if ((op_tag != csg_op_not) && (op_tag != csg_op_is))
    if (rtag == csg_prim)
      surf_right->Print();
    else
      csg_right->print(level + 1);
}

Interval::Interval(Surface *surf1, Surface *surf2, v2 *surf_interval) {
  ival = surf_interval;
  s_in = surf1;
  sout = surf2;
  next = NULL;
}

Interval::~Interval() {
  if (next != NULL)
	delete next;
  if (ival == NULL)
	printf("Warning: ival null!\n");
  else
  	delete ival;
}

void Interval::print() {
	ival->print();
	printf("  ");
	if (next != NULL)
		next->print();
	else
		printf("\n");
}

Interval *csg_not(Interval *i) {
  v2 *itmp;
  real newx, newy;
  Interval *iout, *lastout, *rootout;
  rootout = lastout = NULL;
  Surface *s_in, *sout;

  if (i == NULL)
    {
      itmp = new v2(-BIGREAL, BIGREAL);
      return new Interval(NULL, NULL, itmp);
    }
  if (i->ival->x <= -BIGREAL)
  {
      if (i->ival->y >= BIGREAL) 
	{
		return NULL;
	}
  } else {
	// need -Bigreal, first x interval, i.e. not left real line
      sout = i->s_in;
      itmp = new v2(-BIGREAL, i->ival->x);
      iout = new Interval(NULL, sout, itmp);
      rootout = lastout = iout;
  }
	
  newx = i->ival->y;
  s_in = i->sout;
     
  while ((i = i->next) != NULL)
  {
	newy = i->ival->x;
	sout = i->s_in;
	itmp = new v2(newx, newy);
	if (rootout == NULL)
	{
		iout = new Interval(s_in, sout, itmp);
		lastout = iout;
		rootout = iout;
	} else  {
		iout = new Interval(s_in, sout, itmp);
		lastout->next = iout;
		lastout = iout;
	}
	newx = i->ival->y;
	s_in = i->sout;
	if (newx >= BIGREAL)
		break;
  }
  if (newx < BIGREAL)
    {	
      itmp = new v2(newx, BIGREAL);
      iout = new Interval(s_in, NULL, itmp); 
      if (lastout != NULL)
	  lastout->next = iout;
      else
	{
	  rootout = iout;
	  rootout->next = NULL;
	}
    }
  return rootout;
}

Interval *csg_or(Interval *a, Interval *b) {
  v2 *itmp;
  real newx, newy, ax, ay, bx, by;
  Interval *iout, *lastout, *rootout, *itail;
  rootout = lastout = NULL;
  Surface *s_in, *sout, *new_sin;
  char newinterval;

  newinterval = 'n';

  if (a == NULL)
	return b;
  if (b == NULL)
	return a;
  if (a->ival->x < b->ival->x) {
	newx = a->ival->x;
	s_in = a->s_in;
  } else {
	newx = b->ival->x;
	s_in = b->s_in;
  }

  while (( a != NULL) && (b != NULL))
  {
	ax = a->ival->x;
	ay = a->ival->y;
	bx = b->ival->x;
	by = b->ival->y;

	if (ax > bx)
	{
		if (ax > by)
		{
			itmp = new v2(newx, by);
			sout = b->sout;
			newinterval = 'y';
			b = b->next;
  			if (b == NULL) {
				newx = ax;
				new_sin = a->s_in;
			} else if (a == NULL) {
				newx = bx;
				new_sin = b->s_in;
			} else if (a->ival->x < b->ival->x) {
				newx = a->ival->x;
				new_sin = a->s_in;
  			} else {
				newx = b->ival->x;
				new_sin = b->s_in;
  			}
		} else {
			if (ay > by)
			{
				b = b->next;
			} else {
				a = a->next;
			}
		}
	} else {
		if (bx > ay)
		{
			itmp = new v2(newx, ay);
			sout = a->sout;
			newinterval = 'y';
			a = a->next;
  			if (a == NULL) {
				newx = bx;
				new_sin = b->s_in;
			} else if (b == NULL) {
				newx = ax;
				new_sin = a->s_in;
			} else if (a->ival->x < b->ival->x) {
				newx = a->ival->x;
				new_sin = a->s_in;
  			} else {
				newx = b->ival->x;
				new_sin = b->s_in;
  			}
		} else {
			if (by > ay)
			{
				a = a->next;
			} else {
				b = b->next;
			}
		}
	}
	if (newinterval == 'y') 
	{
		newinterval = 'n';
		if (rootout == NULL)
        	{
               		iout = new Interval(s_in, sout, itmp);
               		lastout = iout;
               		rootout = iout;
        	} else  {
               		iout = new Interval(s_in, sout, itmp);
               		lastout->next = iout;
               		lastout = iout;
        	}
		s_in = new_sin;
	}
  } // while
  if (a == NULL) {
	sout = b->sout;
	itail = b->next;
	b->next = NULL;  // since the tail of b will concatinate the result,
			 // and I delete b when done, I tell b it has no tail
	itmp = new v2(newx, by);
  } else {
	sout = a->sout;
	itail = a->next;
	a->next = NULL;
	itmp = new v2(newx, ay);
  }
if (rootout == NULL)
 {
      	iout = new Interval(s_in, sout, itmp);
    	lastout = iout;
       	rootout = iout;
	lastout->next = itail;
 } else  {
 	iout = new Interval(s_in, sout, itmp);
        lastout->next = iout;
        lastout = iout;
	lastout->next = itail;
 } 
 return rootout;
}

Interval *csg_and(Interval *a, Interval *b) {
  v2 *itmp;
  real newx, newy, ax, ay, bx, by;
  Interval *iout, *lastout, *rootout;
  rootout = lastout = NULL;
  char newinterval = 'n';
  Surface *s_in, *sout;

  while (( a != NULL) && (b != NULL))
  {
	ax = a->ival->x;
	ay = a->ival->y;
	bx = b->ival->x;
	by = b->ival->y;

	if (ax > bx)
	{
		if (ax > by)
		{
			b = b->next;
		} else {
			if (ay > by)
			{
				itmp = new v2(ax, by);
				s_in = a->s_in;
				sout = b->sout;
				newinterval = 'y';
				b = b->next;
			} else {
				itmp = new v2(ax, ay);
				s_in = a->s_in;
				sout = a->sout;
                                newinterval = 'y';
				a = a->next;
			}
		}
	} else {
		if (bx > ay)
		{
			a = a->next;
		} else {
			if (by > ay)
			{
				itmp = new v2(bx, ay);
				s_in = b->s_in;
				sout = a->sout;
                                newinterval = 'y';
				a = a->next;
			} else {
				itmp = new v2(bx, by);
				s_in = b->s_in;
				sout = b->sout;
                                newinterval = 'y';
				b = b->next;
			}
		}
	}
	if (newinterval == 'y') 
	{
		newinterval = 'n';
		if (rootout == NULL)
        	{
               		iout = new Interval(s_in, sout, itmp);
               		lastout = iout;
               		rootout = iout;
        	} else  {
               		iout = new Interval(s_in, sout, itmp);
               		lastout->next = iout;
               		lastout = iout;
        	}
	}
	
  }
  return rootout;
}


Scene::Scene() {
  root = current = NULL;
};


v3 * Scene::Testray(ray *a_ray)
{
  real closest_hit;
  v3 *result;
  Surface *surface, *closest_surface;
  Interval *hitlist, *hitroot;

  hitroot = hitlist = interval(csgnode, a_ray);
  while (hitlist != NULL) {
	if (hitlist->ival->x >= 0.00)
		break;  // find closest interval in front of ray origin
	hitlist = hitlist->next;
  }
  if (hitlist == NULL) {
	result = new v3;
	result->x = result->y = result->z = 0.0;
  } else
    result = Shader(a_ray, hitlist->ival->x, hitlist->s_in);
  delete hitroot;
  delete a_ray;
  return result;
}

real Scene::Feeler(ray *a_ray)
{
  real closest_hit;
  real result;
  Interval *hitlist, *hitroot;

  hitroot = hitlist = interval(csgnode, a_ray);
  while (hitlist != NULL) {
	if (hitlist->ival->x >= 0.00)
		break;  // find closest interval in front of ray origin
	hitlist = hitlist->next;
  }
  if (hitlist == NULL) {
	result = BIGREAL;
  } else
    result = hitlist->ival->x;
  delete hitroot;
  delete a_ray;
  return result;
}

void Scene::Print() {
	csgnode->print(1);
}

Interval *interval(CSGnode *node, ray *r)
{
	Interval *a, *b, *c;
	v2 *va, *vb;

	if (node->ltag == csg_prim)
		if ((va = node->surf_left->interval(r)) == NULL)
			a = NULL;
		else
			a = new Interval(node->surf_left, node->surf_left, va);
	else
		a = interval(node->csg_left, r);
	if ((node->op_tag != csg_op_not) && (node->op_tag != csg_op_is)) {
		if (node->rtag == csg_prim)
			if ((vb = node->surf_right->interval(r)) == NULL)
				b = NULL;
			else
				b = new Interval(node->surf_right, 
					node->surf_right,vb);
		else
			b = interval(node->csg_right, r);
	} else {
		b = NULL;
	}
	switch (node->op_tag) {
		case csg_op_union: c = csg_or(a, b);
					//if (a != NULL)
						//a->print();
					//printf(" or ");
					//if (b != NULL)
						//b->print();
					//printf(" = ");
					//if (c != NULL)
						//c->print();
					//printf(" **\n");
					break;
		case csg_op_intersection: c = csg_and(a, b);
					break;
		case csg_op_not:  c = csg_not(a);
					break;
		case csg_op_is:	  c = a;
					break;
		};
	if (a != NULL) 
		if (c != a)
			delete a;
	if (b != NULL) 
		if (c != b)
			delete b;
	return c;
}



