/* $Id: shader.cc,v 1.6 1997/04/12 22:32:46 sean Exp sean $ */
/*
$Log: shader.cc,v $
Revision 1.6  1997/04/12 22:32:46  sean
*** empty log message ***

// Revision 1.5  1996/05/06  03:57:40  sean
// bump added
//
Revision 1.4  1995/08/03 18:45:43  sean
Diffuse and Reflect working.  Need Specular and Transmit

*/
#include <stdio.h>
#include <math.h>
#include "ray.h"
#include "matrix.h"
#include "surfaces.h"
#include "lights.h"
#include "globals.h"
#include "csg.h"

v3 SpecDiffuse(v3 *surfnormal, Surface *surface, 
                 v3 *intersection, ray *incident_ray, real diffscale, real specscale);
v3 Reflect(v3 *surfnormal, v3 incidentdir);

extern Scene *scene;

v3 * Shader(ray * incident_ray, real photon_distance, Surface *surface)
{
  v3 *intersection, *surfnormal;
  v3 diffuse, specular, reflect, refldir;
  v2 surfpoint;
  v3 *result, *reflcol;
  v3 color;
  ray *reflray;

  result = new v3;
  intersection = new v3;
  *intersection = 
    v3_scalar_mul(photon_distance, incident_ray->direction) +
      incident_ray->origin;
  
  surfnormal = surface->Normal(intersection);
  if (v3_dot(*surfnormal, incident_ray->direction) > 0.0)
	*surfnormal = - *surfnormal;
  *surfnormal = surfnormal->unit();

  // the result is ambiant + diffuse + specular + transmitted + reflected
  //           diffuse = lightdir . surfnormal
  //          specular = ( (lightdir - surfnormal ) . ( incidentdir - surfnormal)) ^ X
  //          reflected= spawn ray   refldir = see below

  diffuse = SpecDiffuse(surfnormal, surface, intersection, incident_ray,
		surface->surface_props->diffuse, surface->surface_props->specular);
  if ((surface->surface_props->refl != 0.0) && 
		(incident_ray->spawn_level < 50)) {
	refldir = Reflect(surfnormal,  incident_ray->direction);	
	reflray = new ray;
	reflray->origin = *intersection + v3_scalar_mul(0.0001, *surfnormal);
	reflray->direction = refldir;
	reflray->spawn_level = incident_ray->spawn_level + 1;
	reflcol = scene->Testray(reflray);
	reflect = v3_scalar_mul(surface->surface_props->refl, *reflcol);
	delete reflcol;
  } else 
	reflect = v3(0.,0.,0.);
  *result = diffuse;
  *result = *result + reflect;

  delete surfnormal;
  delete intersection;
  return result;
}

v3 Reflect(v3 *surfnormal, v3 incidentdir)
// returns ray reflected about surfnormal: twice the projection of i onto n
// minus i. vectors assumed unit.
{
	real dot;
	v3 negincident, rdir;

	negincident = - incidentdir;

	dot = v3_dot( *surfnormal, negincident);
	
	rdir = v3_scalar_mul( (2.0 * dot) , *surfnormal) - negincident;
	return rdir.unit();
}

v3 SpecDiffuse(v3 *surfnormal, Surface *surface, 
                 v3 *intersection, ray *incident_ray, real diffscale, real specscale)
{
  LightList *lightlist = new LightList(global_lightlist);
  LightSource *lightsource;
  v3 result, light_unit_normal, *spectrum, *n, dresult, sresult,color, refldir;
  ray *shadow_feeler;
  real light_distance, feeler_distance, specindex;
  v2 surfpoint;
  
  result = v3(0., 0., 0.);

  if ((diffscale == 0.0) && (specscale == 0.0))
	return result;

  sresult = v3(0., 0., 0.);
  dresult = v3(0., 0., 0.);
  surfpoint = surface->R2(*intersection);
  color = surface->color(surfpoint);
  lightsource = lightlist->First();
  // delete lightsourse->Normal
  while (lightsource != NULL) { //do shadow stuff here
    n = lightsource->Normal(intersection);
    light_unit_normal = n->unit();
    shadow_feeler = new ray;
    shadow_feeler->direction = light_unit_normal;
    shadow_feeler->origin = *intersection +  v3_scalar_mul(0.0001, *surfnormal);
    feeler_distance = scene->Feeler(shadow_feeler);
    if (feeler_distance * feeler_distance > v3_dot(*n , *n)) {
	if (diffscale != 0.00)
    		dresult = dresult + v3_scalar_mul(
				 v3_dot(light_unit_normal, *surfnormal), 
			    *(lightsource->spectrum)).positive() * color;
	if (specscale != 0.00) {
		refldir = Reflect(surfnormal, *n).unit();	
		specindex = v3_dot(incident_ray->direction, refldir); 
		if (specindex > 0.0) {
                	specindex = pow(specindex, 12.0);
			sresult = sresult + v3_scalar_mul(specindex,
						*(lightsource->spectrum));
		}
	}
    }
    lightsource = lightlist->Next();
    delete n;
  }
  delete lightlist;
  result = v3_scalar_mul(diffscale, dresult) + 
	   v3_scalar_mul(specscale, sresult) + (AMBIENT * color);
  return result;
}

