/* $Id: surfaces.cc,v 1.8 1997/04/12 22:32:46 sean Exp sean $ */
/*
$Log: surfaces.cc,v $
Revision 1.8  1997/04/12 22:32:46  sean
*** empty log message ***

// Revision 1.7  1996/05/06  03:56:54  sean
// bumps on planes
//
Revision 1.6  1995/08/03 18:45:43  sean
Spheres and Planes.  re-think pointer/object return policy.
Add cones, cylinders soon.

*/
#include <stdio.h>
#include <math.h>
#include "globals.h"
#include "surfaces.h"
#include "shader.h"

Sphere::Sphere(v3 newcenter, real newradius, v3 newpole, v3 newequator)
{
  pole = new v3;
  equator = new v3;
  center = new v3;
  *center = newcenter;
  radius = newradius;
  radius_pow_2 = radius * radius;
  *pole = newpole.unit();
// to insure the equator is 90deg from the pole, and sp x se
// is perpendicular to both, I cross product everyone.
  *equator = newequator.unit();
  spxse = new v3;
  *spxse = v3_cross( *pole, *equator).unit();
  *equator = v3_cross( *spxse, *pole).unit();
}

Plane::Plane(v3 neworigin, v3 newu, v3 newv)
{
  origin = new v3;
  u = new v3;
  v = new v3;
  unitnormal = new v3;

  *origin = neworigin;
  *u = newu;
  *v = newv;
  *unitnormal = (v3_cross(newu, newv)).unit();
  // let O be the origin of the plane, and N be the normal and P be some
  // point on the plane.  Then the vector P - O must be perpendicular to N,
  // or the cosine of the angle between them must be zero, or (P - O) . N =
  // |P - O| * |N| * cos(angle) = |P - O| * |N| * 0 = 0.
  // This equation is used in the calculation of a,b,c and d in ax+by+cz=d
  // as the plane equation.
  // (p.x - o.x) * n.x + (p.y - o.y) * n.y + (p.z - o.z) * n.z = 0
  // n.x * p.x + n.y * p.y + n.z * p.z = n.x * o.x + n.y * o.y + n.z * o.z
  // gosh, N . P = N . O  irrelavant, I think, but interesting all the same
  // a = n.x, b = n.y, c = n.z, d = N . O
  a = unitnormal->x;
  b = unitnormal->y;
  c = unitnormal->z;
  d = v3_dot(*unitnormal, *origin);
}

Mesh::Mesh( Plane *newplane, PPMrast *newpgm)
{
	//4/11/97 What am I doing here?
	baseplane = newplane;
	int i,j;
	real scale;
	MeshNode *newnode, *oldnode;
	v3 tempo, tempu, tempv, du, dv;
	
	meshlist = NULL;
	scale = 10.0;
	rows = newpgm->width - 1;
	cols = newpgm->height - 1;
	du = v3_scalar_mul(1.0 / ((real) rows + 1.0) , *baseplane->u);
	dv = v3_scalar_mul(1.0 / ((real) cols + 1.0) , *baseplane->v);
	for (j = 0 ; j < cols ; j++)
	for (i = 0 ; i < rows ; i++)
	{
	  newnode = new(MeshNode);
	  tempo = *baseplane->origin +
		  v3_scalar_mul( ((real) i)  / ( (real) rows + 1.0 ), du)  +
		  v3_scalar_mul( ((real) j)  / ( (real) cols + 1.0 ), dv)  +
		  v3_scalar_mul( ((real) newpgm->getpixel(v2(i,j)).x) * scale , *baseplane->unitnormal);
	  tempu = v3_scalar_mul( ((real) i + 1)  / ( (real) rows + 1.0 ), du)  +
		  v3_scalar_mul( ((real) j)  / ( (real) cols + 1.0 ), dv)  +
		  v3_scalar_mul( ((real) newpgm->getpixel(v2(i + 1,j)).x) * scale , *baseplane->unitnormal);
	  tempv = v3_scalar_mul( ((real) i)  / ( (real) rows + 1.0 ), du)  +
		  v3_scalar_mul( ((real) j + 1)  / ( (real) cols + 1.0 ), dv)  +
		  v3_scalar_mul( ((real) newpgm->getpixel(v2(i,j + 1)).x) * scale , *baseplane->unitnormal);

	  newnode->sex = 'M';
	  newnode->row = i;
	  newnode->col = j;
	  newnode->plane = new Plane( tempo, tempu, tempv );
	  newnode->next = meshlist;
	  meshlist = newnode;
	}
}
	

Cylinder::Cylinder(v3 newcenter, v3 newaxis, v3 newequator, real newradius)
{
  center = new v3;
  axis = new v3;
  equator = new v3;
  cxa = new v3;

  radius = newradius;
  *center = newcenter;
  *axis = newaxis;
  *equator = newequator;
  *cxa = v3_cross(*center, *axis);
  aa = v3_dot(*axis, *axis);
  rr = radius * radius;
  rraa = radius * radius * aa;
}

void Sphere::Print()
{
  printf("Sphere at ");
  center->print();
  printf(" with radius %3.2f\n", radius);
}

void Plane::Print()
{
  printf("Plane at ");
  origin->print();
  printf(" with normal ");
  unitnormal->print();
  printf("\n\tAka: %3.2f x + %3.2f y + %3.2f z = %3.2f \n", a, b, c, d);
}

void Mesh::Print()
{
  printf("Mesh Base: ");
  baseplane->Print();
}

void Cylinder::Print()
{
  printf("Cylinder at ");
  center->print();
  printf(" with axis ");
  axis->print();
  printf(" with radius %3.2f\n", radius);
}


v3 * Sphere::Normal(v3 *intersection)
{
  v3 *result;

  result = new v3;

  *result = *intersection - *center;
  return result;
}

v3 * Plane::Normal(v3 *intersection)
{
  v3 *result;

  result = new v3;

  *result = *unitnormal;
  return result;
}

v3 * Cylinder::Normal(v3 *intersection)
{
  v3 *result;
  result = new v3;

  *result = v3_cross(v3_cross(*center - *intersection, *axis), *axis);
  return result;
}

v2 Sphere::R2(v3 hit)
{
  v3 sn;
  real theta, phi, dot;

  sn = (hit - *center).unit();
  dot = - v3_dot(sn, *pole);
  if ((dot > 1.0) || (dot < -1.0))
	dot = 1.0;
  theta = acos(dot);
  if (theta != 0.0) {
	dot = v3_dot( *equator, sn) / sin(theta);
	if ((dot > 1.0) || (dot < -1.0))
		dot = 1.0;
  	phi = acos(dot) / (2.0 * M_PI);
  	phi = (v3_dot(*spxse, sn) < 0.0) ?  phi : 1.0 - phi;
  } else {
	phi = 0.0;
  }
  return v2(phi, theta / M_PI);
}

v3 Sphere::R3(v2 cartesian)
{
  v3 result;
//
// *equator, *pole, *center, and *spxse are all part of the fun
// result = pole component + equator component
//  pole component = pole * [radius * cos( (1 - cartesian.y) * M_PI) 
//  eqator component = [ r * cos (2 M_PI cartesian.x) ] * equator +
//		       [ r * sin (2 M_PI cartesian.x) ] * spxse
//  r is the radius of the uh, latitue cirle that the cartesian is on
//  i.e. at the pole, the r is zero, and at the equator, r is the radius
//  of the sphere. clear?
// common factors: 2 MP_I cartesian.x
//		   (1.0 - cartesian.y) * M_PI

  real mpcartx, r, cy;

  mpcartx = 2.0 * M_PI * cartesian.x;
  cy = (1.0 - cartesian.y) * M_PI;
  r = radius * sin(cy);

  result = v3_scalar_mul( (radius * cos(cy) ), *pole) +
	   v3_scalar_mul( (r * cos(mpcartx)), *equator) +
	   - v3_scalar_mul( (r * sin(mpcartx)), *spxse) + *center;

  return result;
}

v2 Plane::R2(v3 hit)
{
  v2 result;
  m2 a;

  a = m2(u->x, v->x, u->y, v->y);
  if (a.det() != 0.0) {
    result = a.inv() * v2(hit.x - origin->x, hit.y - origin->y);
  } else {
      a = m2(u->x, v->x, u->z, v->z);
      if (a.det() != 0.0) {
	result = a.inv() * v2(hit.x - origin->x, hit.z - origin->z);
      } else {
	a = m2(u->y, v->y, u->z, v->z);
	result = a.inv() * v2(hit.y - origin->y, hit.z - origin->z);
      }
  }
  return result;
}

v3 Plane::R3(v2 cartesian)
{
  v3 result;

  result = *origin + v3_scalar_mul(cartesian.x, *u) + v3_scalar_mul(cartesian.y, *v);
  return result;
}

v2 Cylinder::R2(v3 hit)
{
  v2 result;
  v3 hc, p, i;
  // hc is the vector from the hit to the center of the cylender
  // p is the point on the axis closest to the hit
  // i is the vector from hit to p.  
  real k, c1, hchc;
  // k is the number of axis vectors needed to get from the center to p
  // c1 is the arc cosine of the angle between i and the equator
  // The function returns the pair(k, angle of i and the equator / 2pi)

 hc = *center - hit;
 hchc = v3_dot(hc, hc) - rr;
 if (hchc < 0.0)
 	k = 0;
 else
 	k = sqrt(hchc / aa);
 p = *center + v3_scalar_mul(k, *axis);
 i = hit - p;
 c1 = acos(v3_dot(equator->unit(), i.unit())) / M_PI;
 if (v3_dot(v3_cross(*equator, i), v3_cross(*axis, i)) > 0.0)
	c1 = 1.0 - c1;
 result = v2(c1, k);
 return result;
}
  

Surface::Surface(surfaceTag new_tag, void *arg1, SurfProps *defsurfprops)
{
  tag = new_tag;
  switch (tag) {
  case Sphere_tag:
    sphere = (Sphere *) arg1;
    break;
  case Cylinder_tag:
    cylinder = (Cylinder *) arg1;
    break;
  case Plane_tag:
    plane = (Plane *) arg1;
    break;
  default:
    break;
  }; //switch
  surface_props = defsurfprops;
}

v3 * Surface::Normal(v3 *intersection)
{
  v3 *result;
 
  	switch (tag) {
  	case Sphere_tag:
    		result = sphere->Normal(intersection);
    		break; //sphere
  	case Cylinder_tag:
    		result = cylinder->Normal(intersection);
    		break;
  	case Plane_tag:
    		result = plane->Normal(intersection);
    		break;
  	case Mesh_tag:
    		//result = mesh->Normal(intersection);
    		break;
  	default:
    		break;
  	} //switch
  if (surface_props->bumptag == Pixmap_bump)
  {
	v2 ii2, uu2, vv2;
	v3 n, ii3, uu3, vv3, vvv3, uuu3, bi, bu, bv,nn;
	real ci, cu, cv, dot;	
	
	nn = *result;

  	switch (tag) {
  	case Sphere_tag:
		ii2 = sphere->R2(*intersection);
		uu2 = ii2 + v2(2.0 * surface_props->ppmbump->inv_width, 0.0);
		vv2 = ii2 + v2(0.0, 2.0 * surface_props->ppmbump->inv_height);
		ci = surface_props->Bump(ii2);
		cu = surface_props->Bump(uu2);
		cv = surface_props->Bump(vv2);
		//ii3 = plane->R3(ii2);
		//(*intersection - ii3).print();
		uu3 = sphere->R3(uu2);
		vv3 = sphere->R3(vv2);
		bi = *intersection + v3_scalar_mul(ci * .1 , nn);
		bu = v3_scalar_mul(surface_props->ppmbump->width,
			(uu3 - bi) + v3_scalar_mul(cu  * .1, nn));
		bv = v3_scalar_mul(surface_props->ppmbump->height,
			(vv3 - bi) + v3_scalar_mul(cv  * .1, nn));
		*result = v3_cross(bu, bv);
		if (v3_dot(*result, *result) <= 0.000001)
			*result = nn;
    		break;
  	case Cylinder_tag:
    		result = cylinder->Normal(intersection);
    		break;
  	case Plane_tag:
		ii2 = plane->R2(*intersection);
		if (ii2.x > 0.0) 
			uu2 = ii2 + v2(2.0 * surface_props->ppmbump->inv_width, 0.0);
		else 
			uu2 = ii2 - v2(2.0 * surface_props->ppmbump->inv_width, 0.0);
		if (ii2.y > 0.0) 
			vv2 = ii2 + v2(0.0, 2.0 * surface_props->ppmbump->inv_height);
		else 
			vv2 = ii2 - v2(0.0, 2.0 * surface_props->ppmbump->inv_height);
		
		ci = surface_props->Bump(ii2);
		cu = surface_props->Bump(uu2);
		cv = surface_props->Bump(vv2);
		//ii3 = plane->R3(ii2);
		//(*intersection - ii3).print();
		uu3 = plane->R3(uu2);
		vv3 = plane->R3(vv2);
		bi = *intersection + v3_scalar_mul(ci * .1 , nn);
		bu = v3_scalar_mul(surface_props->ppmbump->width,
			(uu3 - bi) + v3_scalar_mul(cu  * .1, nn));
		bv = v3_scalar_mul(surface_props->ppmbump->height,
			(vv3 - bi) + v3_scalar_mul(cv  * .1, nn));
		*result = v3_cross(bu, bv);
		if (v3_dot(*result, *result) <= 0.000001)
			*result = nn;
//		dot = v3_dot(result->unit(), nn);
//		printf("%3.2f %3.2f %3.2f\n",dot, v3_dot(nn,nn), v3_dot(*result,
//			*result));
    		break;
	case Mesh_tag:
		fprintf(stderr, "Bumpy meshes???\n");
  	default:
    		break;
  	} //switch
	
  }
 

  return result;
}  

v2 Surface::R2(v3 intersection)
{
  v2 result;
  
  switch (tag) {
  case Sphere_tag:
    result = sphere->R2(intersection);
    break; //sphere
  case Cylinder_tag:
    result = cylinder->R2(intersection);
    break;
  case Plane_tag:
    result = plane->R2(intersection);
    break;
  case Mesh_tag:
    //result = mesh->R2(intersection);
    break;
  default:
    break;
  } //switch
  return result;
}  

v2 *Plane::interval(ray *testme)
{
	v2 *result = NULL;
	real divisor, tray;

  // for testme to hit the plane, a * (p.dx * t + p.ox ) + b ... + c ... = d
  // or (a * p.dx + b * p.dy + c * p.dz) * t = d - (a * p.ox + b ... + c ...)

  divisor = a * testme->direction.x + b * testme->direction.y + 
    c * testme->direction.z;
  if (divisor == 0.0)  // missed
	{
		if (v3_dot(testme->origin - *origin, *unitnormal) > 0.0)
	 		{
			//	printf("missed plane");
			}
		else
			{
			//	printf("HIT");
				result = new v2(-BIGREAL, BIGREAL);
			}
		return result;
	}
// hit  possible negative, but algorithm takes care of that.
     tray = (d - a * testme->origin.x - b * testme->origin.y -
		    c * testme->origin.z) / divisor;
     if ((tray < BIGREAL) && (tray > -BIGREAL)) {
     	if (v3_dot(*unitnormal, (testme->origin - *origin)) >= 0.0) {
		if (tray >= 0.0)
			result = new v2(tray, BIGREAL);
		else
			result = new v2(-BIGREAL, tray);
     	} else {
		if (tray >= 0.0)
			result = new v2(-BIGREAL, tray);
		else
			result = new v2(tray, BIGREAL);
	}
	//result->print();
	//printf("   [%3.2g %3.2g %3.2g] = [%3.2g]\n", a, b, c, d);
    }
    return result;
}

v2 *Mesh::interval(ray *r)
{
	//4/11/97 What am I doing here?
	// how about what "an introduction to ray tracing" says?
	// Ahem, make it 2d-ish.  Really, really define your metrics.
	// I mean, where triangles diagonalize.  Let's keep those normals
	// up! 5/2/97
	int i,j;
	MeshNode *newnode, *oldnode;
	v3 tempo, tempu, tempv, du, dv;
	v2 *result, *contender, coord;

	
	result = NULL;

	newnode = meshlist;
	while (newnode != NULL)
	{
		contender = newnode->plane->interval(r);
		if (result == NULL)
			result = contender;
		else
		{
			if ( contender->x < result->x)
				result = contender;
		}
	}
	return result;
}

v2 *Sphere::interval(ray *r)
{
	real tmpx, tmpy, tmpz, a, b, c, radicle, q, r1, r2;
	v2 *result;
	
	a = r->direction.x * r->direction.x +
	r->direction.y * r->direction.y +
	r->direction.z * r->direction.z;
	if (a == 0.0)
		return NULL;
	tmpx = r->origin.x - center->x;
	tmpy = r->origin.y - center->y;
	tmpz = r->origin.z - center->z;

	b = 2.0 * (r->direction.x * tmpx + r->direction.y * tmpy + 
			r->direction.z * tmpz);
	c = tmpx * tmpx + tmpy * tmpy + tmpz * tmpz - radius_pow_2;
	radicle = b * b - 4 * a * c;
	if (radicle < 0.0)
		return NULL;
	if (b >= 0.0)
		q = -0.5 * (b + sqrt(radicle));
	else
		q = -0.5 * (b - sqrt(radicle));
	if (q == 0.0)
		return NULL;
	r1 = q / a;
	r2 = c / q;
	if (r1 > r2)
		result = new v2(r2, r1);
	else
		result = new v2(r1, r2);
	return result;
}

v2 *Cylinder::interval(ray *r)
{
	v3 vtmp;
  	v2 *result;
	v3 axd;
 	real a, b, c, radicle, q, r1, r2;

	axd = v3_cross(*axis, r->direction);
	vtmp = *cxa + v3_cross(*axis, r->origin);
	a = v3_dot(axd, axd);
        if (a == 0.0)
		return NULL;
        b = 2.0 * (v3_dot(axd, v3_cross(*axis, (r->origin - *center))));
	c = v3_dot(vtmp, vtmp) - rraa;
	
	radicle = b * b - 4 * a * c;
        if (radicle < 0.0)
                return NULL;
        if (b >= 0.0)
                q = -0.5 * (b + sqrt(radicle));
        else
                q = -0.5 * (b - sqrt(radicle));
	if (q == 0.0)
		return NULL;
        r1 = q / a;
        r2 = c / q;
        if (r1 > r2)
                result = new v2(r2, r1);
        else
                result = new v2(r1, r2);
	return result;
}

v3 Surface::color(v2 surfpoint)
{
  v3 result;

  result = surface_props->Color(surfpoint);
  return result;
}

v2 *Surface::interval(ray *r)
{
	
  v2 *result;

  switch (tag) {
  case Sphere_tag:
    result = sphere->interval(r);
    break; //sphere
  case Cylinder_tag:
    result = cylinder->interval(r);
    break;
  case Plane_tag:
    result = plane->interval(r);
    break;
  case Mesh_tag:
    result = mesh->interval(r);
    break;
  default:
    break;
  } //switch
  return result;
}

void Surface::Print() {
  switch (tag) {
  case Sphere_tag:
      sphere->Print();
      break;
    case Cylinder_tag:
      cylinder->Print();
      break;
    case Plane_tag:
      plane->Print();
      break;
    case Mesh_tag:
      mesh->Print();
      break;
    default:
      break;
    } //switch
}

